@@ -1,11 +1,12 @@
 /*
  * The initial "port" of dwm to curses was done by
- * (c) 2007-2012 Marc Andre Tanner <mat at brain-dump dot org>
+ *
+ * © 2007-2013 Marc André Tanner <mat at brain-dump dot org>
  *
  * It is highly inspired by the original X11 dwm and
  * reuses some code of it which is mostly
  *
- * (c) 2006-2007 Anselm R. Garbe <garbeam at gmail dot com>
+ * © 2006-2007 Anselm R. Garbe <garbeam at gmail dot com>
  *
  * See LICENSE for details.
  */
@@ -13,14 +14,10 @@
 #define _GNU_SOURCE
 #include <sys/stat.h>
 #include <sys/ioctl.h>
-#ifdef _AIX
-# include <fcntl.h>
-#else
-# include <sys/fcntl.h>
-#endif
 #include <sys/wait.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <fcntl.h>
 #include <curses.h>
 #include <stdio.h>
 #include <stdarg.h>
@@ -44,7 +41,7 @@ int ESCDELAY;
 #endif
 
 typedef struct {
-	double mfact;
+	float mfact;
 	int history;
 	int w;
 	int h;
@@ -61,8 +58,8 @@ struct Client {
 	WINDOW *window;
 	Vt *term;
 	const char *cmd;
-	char title[256];
-	uint8_t order;
+	char title[255];
+	int order;
 	pid_t pid;
 	int pty;
 	unsigned short int id;
@@ -92,13 +89,13 @@ typedef struct {
 #endif
 #define CTRL_ALT(k) ((k) + (129 - 'a'))
 
-#define MAX_ARGS 2
+#define MAX_ARGS 3
 
 typedef struct {
 	void (*cmd)(const char *args[]);
 	/* needed to avoid an error about initialization
 	 * of nested flexible array members */
-	const char *args[MAX_ARGS + 1];
+	const char *args[MAX_ARGS];
 } Action;
 
 typedef struct {
@@ -147,6 +144,7 @@ typedef struct {
 
 /* commands for use by keybindings */
 static void create(const char *args[]);
+static void copymode(const char *args[]);
 static void escapekey(const char *args[]);
 static void focusn(const char *args[]);
 static void focusnext(const char *args[]);
@@ -155,6 +153,7 @@ static void focusprev(const char *args[]);
 static void focusprevnm(const char *args[]);
 static void killclient(const char *args[]);
 static void lock(const char *key[]);
+static void paste(const char *args[]);
 static void quit(const char *args[]);
 static void redraw(const char *args[]);
 static void scrollback(const char *args[]);
@@ -179,6 +178,8 @@ static void resize(Client *c, int x, int y, int w, int h);
 extern Screen screen;
 static unsigned int waw, wah, wax, way;
 static Client *clients = NULL;
+static char *title;
+#define COLOR(fg, bg) COLOR_PAIR(vt_color_reserve(fg, bg))
 
 #include "config.h"
 
@@ -191,7 +192,8 @@ static Layout *layout = layouts;
 static StatusBar bar = { -1, BAR_POS, 1 };
 static CmdFifo cmdfifo = { -1 };
 static const char *shell;
-static bool running = true;
+static char *copybuf;
+static volatile sig_atomic_t running = true;
 static bool runinall = false;
 
 static void
@@ -231,7 +233,6 @@ drawbar() {
 		return;
 	curs_set(0);
 	attrset(BAR_ATTR);
-	wcolor_set(stdscr, vt_color_get(BAR_FG, BAR_BG), NULL);
 	mvaddch(bar.y, 0, '[');
 	if (mbstowcs(wbuf, bar.text, sizeof bar.text) == (size_t)-1)
 		return;
@@ -258,13 +259,7 @@ draw_border(Client *c) {
 	char t = '\0';
 	int x, y, maxlen;
 
-	if (sel == c || (runinall && !c->minimized)) {
-		wattrset(c->window, SELECTED_ATTR);
-		wcolor_set(c->window, vt_color_get(SELECTED_FG, SELECTED_BG), NULL);
-	} else {
-		wattrset(c->window, NORMAL_ATTR);
-		wcolor_set(c->window, vt_color_get(NORMAL_FG, NORMAL_BG), NULL);
-	}
+	wattrset(c->window, (sel == c || (runinall && !c->minimized)) ? SELECTED_ATTR : NORMAL_ATTR);
 	getyx(c->window, y, x);
 	curs_set(0);
 	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
@@ -333,7 +328,6 @@ static void
 arrange() {
 	clear_workspace();
 	attrset(NORMAL_ATTR);
-	color_set(vt_color_get(NORMAL_FG, NORMAL_BG), NULL);
 	layout->arrange();
 	wnoutrefresh(stdscr);
 	draw_all(true);
@@ -341,19 +335,17 @@ arrange() {
 
 static void
 attach(Client *c) {
-	uint8_t order;
 	if (clients)
 		clients->prev = c;
 	c->next = clients;
 	c->prev = NULL;
 	clients = c;
-	for (order = 1; c; c = c->next, order++)
-		c->order = order;
+	for (int o = 1; c; c = c->next, o++)
+		c->order = o;
 }
 
 static void
 attachafter(Client *c, Client *a) { /* attach c after a */
-	uint8_t o;
 	if (c == a)
 		return;
 	if (!a)
@@ -365,7 +357,7 @@ attachafter(Client *c, Client *a) { /* attach c after a */
 		c->next = a->next;
 		c->prev = a;
 		a->next = c;
-		for (o = a->order; c; c = c->next)
+		for (int o = a->order; c; c = c->next)
 			c->order = ++o;
 	}
 }
@@ -385,12 +377,22 @@ detach(Client *c) {
 	c->next = c->prev = NULL;
 }
 
+static void
+settitle(Client *c) {
+	char *t = title;
+	if (!t && sel == c && *c->title)
+		t = c->title;
+	if (t)
+		printf("\033]0;%s\007", t);
+}
+
 static void
 focus(Client *c) {
 	Client *tmp = sel;
 	if (sel == c)
 		return;
 	sel = c;
+	settitle(c);
 	if (tmp) {
 		draw_border(tmp);
 		wrefresh(tmp->window);
@@ -403,12 +405,11 @@ focus(Client *c) {
 
 static void
 applycolorrules(Client *c) {
-	unsigned int i;
-	unsigned attrs = A_NORMAL;
-	short fg = -1, bg = -1;
-	const ColorRule *r;
+	const ColorRule *r = colorrules;
+	short fg = r->fg, bg = r->bg;
+	unsigned attrs = r->attrs;
 
-	for (i = 0; i < countof(colorrules); i++) {
+	for (unsigned int i = 1; i < countof(colorrules); i++) {
 		r = &colorrules[i];
 		if (strstr(c->title, r->title)) {
 			attrs = r->attrs;
@@ -429,9 +430,16 @@ term_event_handler(Vt *term, int event, void *event_data) {
 		if (event_data)
 			strncpy(c->title, event_data, sizeof(c->title) - 1);
 		c->title[event_data ? sizeof(c->title) - 1 : 0] = '\0';
+		settitle(c);
 		draw_border(c);
 		applycolorrules(c);
 		break;
+	case VT_EVENT_COPY_TEXT:
+		if (event_data) {
+			free(copybuf);
+			copybuf = event_data;
+		}
+		break;
 	}
 }
 
@@ -515,14 +523,11 @@ sigchld_handler(int sig) {
 			c->died = true;
 	}
 
-	signal(SIGCHLD, sigchld_handler);
-
 	errno = errsv;
 }
 
 static void
 sigwinch_handler(int sig) {
-	signal(SIGWINCH, sigwinch_handler);
 	screen.need_resize = true;
 }
 
@@ -552,11 +557,12 @@ static void
 resize_screen() {
 	struct winsize ws;
 
-	if (ioctl(0, TIOCGWINSZ, &ws) == -1)
-		return;
-
-	screen.w = ws.ws_col;
-	screen.h = ws.ws_row;
+	if (ioctl(0, TIOCGWINSZ, &ws) == -1) {
+		getmaxyx(stdscr, screen.h, screen.w);
+	} else {
+		screen.w = ws.ws_col;
+		screen.h = ws.ws_row;
+	}
 
 	debug("resize_screen(), w: %d h: %d\n", screen.w, screen.h);
 
@@ -644,11 +650,17 @@ setup() {
 	mouse_setup();
 	raw();
 	vt_init();
-	getmaxyx(stdscr, screen.h, screen.w);
+	vt_set_keytable(keytable, countof(keytable));
 	resize_screen();
-	signal(SIGWINCH, sigwinch_handler);
-	signal(SIGCHLD, sigchld_handler);
-	signal(SIGTERM, sigterm_handler);
+	struct sigaction sa;
+	sa.sa_flags = 0;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_handler = sigwinch_handler;
+	sigaction(SIGWINCH, &sa, NULL);
+	sa.sa_handler = sigchld_handler;
+	sigaction(SIGCHLD, &sa, NULL);
+	sa.sa_handler = sigterm_handler;
+	sigaction(SIGTERM, &sa, NULL);
 }
 
 static void
@@ -679,7 +691,11 @@ destroy(Client *c) {
 
 static void
 cleanup() {
+	while (clients)
+		destroy(clients);
+	vt_shutdown();
 	endwin();
+	free(copybuf);
 	if (bar.fd > 0)
 		close(bar.fd);
 	if (bar.file)
@@ -690,6 +706,14 @@ cleanup() {
 		unlink(cmdfifo.file);
 }
 
+static char *getcwd_by_pid(Client *c) {
+	if (!c)
+		return NULL;
+	char buf[32];
+	snprintf(buf, sizeof buf, "/proc/%d/cwd", c->pid);
+	return realpath(buf, NULL);
+}
+
 /* commands for use by keybindings */
 static void
 create(const char *args[]) {
@@ -699,7 +723,7 @@ create(const char *args[]) {
 	const char *cmd = (args && args[0]) ? args[0] : shell;
 	const char *pargs[] = { "/bin/sh", "-c", cmd, NULL };
 	c->id = ++cmdfifo.id;
-	char buf[8];
+	char buf[8], *cwd = NULL;
 	snprintf(buf, sizeof buf, "%d", c->id);
 	const char *env[] = {
 		"DVTM", VERSION,
@@ -707,14 +731,27 @@ create(const char *args[]) {
 		NULL
 	};
 
-	c->window = newwin(wah, waw, way, wax);
-	c->term = vt_create(screen.h - 1, screen.w, screen.history);
+	if (!(c->window = newwin(wah, waw, way, wax))) {
+		free(c);
+		return;
+	}
+
+	if (!(c->term = vt_create(screen.h - 1, screen.w, screen.history))) {
+		delwin(c->window);
+		free(c);
+		return;
+	}
+
 	c->cmd = cmd;
 	if (args && args[1]) {
 		strncpy(c->title, args[1], sizeof(c->title) - 1);
 		c->title[sizeof(c->title) - 1] = '\0';
 	}
-	c->pid = vt_forkpty(c->term, "/bin/sh", pargs, env, &c->pty);
+	if (args && args[2])
+		cwd = !strcmp(args[2], "$CWD") ? getcwd_by_pid(sel) : (char*)args[2];
+	c->pid = vt_forkpty(c->term, "/bin/sh", pargs, cwd, env, &c->pty);
+	if (args && args[2] && !strcmp(args[2], "$CWD"))
+		free(cwd);
 	vt_set_data(c->term, c);
 	vt_set_event_handler(c->term, term_event_handler);
 	c->w = screen.w;
@@ -729,6 +766,17 @@ create(const char *args[]) {
 	arrange();
 }
 
+static void
+copymode(const char *args[]) {
+	if (!sel)
+		return;
+	vt_copymode_enter(sel->term);
+	if (args[0]) {
+		vt_copymode_keypress(sel->term, args[0][0]);
+		draw(sel);
+	}
+}
+
 static void
 escapekey(const char *args[]) {
 	int key;
@@ -848,6 +896,12 @@ lock(const char *args[]) {
 	arrange();
 }
 
+static void
+paste(const char *args[]) {
+	if (sel && copybuf)
+		vt_write(sel->term, copybuf, strlen(copybuf));
+}
+
 static void
 quit(const char *args[]) {
 	cleanup();
@@ -856,6 +910,8 @@ quit(const char *args[]) {
 
 static void
 redraw(const char *args[]) {
+	for (Client *c = clients; c; c = c->next)
+		vt_dirty(c->term);
 	wrefresh(curscr);
 	resize_screen();
 	draw_all(true);
@@ -893,14 +949,14 @@ setlayout(const char *args[]) {
 
 static void
 setmfact(const char *args[]) {
-	double delta;
+	float delta;
 
 	if (isarrange(fullscreen) || isarrange(grid))
 		return;
 	/* arg handling, manipulate mfact */
 	if (args[0] == NULL)
 		screen.mfact = MFACT;
-	else if (1 == sscanf(args[0], "%lf", &delta)) {
+	else if (1 == sscanf(args[0], "%f", &delta)) {
 		if (args[0][0] == '+' || args[0][0] == '-')
 			screen.mfact += delta;
 		else
@@ -1065,7 +1121,7 @@ handle_cmdfifo() {
 				bool quote = false;
 				int argc = 0;
 				/* XXX: initializer assumes MAX_ARGS == 2 use a initialization loop? */
-				const char *args[MAX_ARGS] = { NULL, NULL}, *arg;
+				const char *args[MAX_ARGS] = { NULL, NULL, NULL}, *arg;
 				/* if arguments were specified in config.h ignore the one given via
 				 * the named pipe and thus skip everything until we find a new line
 				 */
@@ -1086,17 +1142,14 @@ handle_cmdfifo() {
 						/* remove the escape character '\\' move every
 						 * following character to the left by one position
 						 */
-						switch (*(++p)) {
+						switch (p[1]) {
 							case '\\':
 							case '\'':
 							case '\"': {
-								char *t = p;
-								for (;;) {
-									*(t - 1) = *t;
-									if (*t++ == '\0')
-										break;
-								}
-								p -= 2;
+								char *t = p+1;
+								do {
+									t[-1] = *t;
+								} while (*t++);
 							}
 						}
 						break;
@@ -1119,12 +1172,14 @@ handle_cmdfifo() {
 
 							while (*p == ' ')
 								++p;
-							arg = p;
+							arg = p--;
 						}
 						break;
 					}
 
 					if (c == '\n' || *p == '\n') {
+						if (!*p)
+							p++;
 						debug("execute %s", s);
 						for(int i = 0; i < argc; i++)
 							debug(" %s", args[i]);
@@ -1211,10 +1266,8 @@ open_or_create_fifo(const char *name, const char **name_created) {
 static void
 usage() {
 	cleanup();
-	eprint("usage: dvtm [-v] [-m mod] [-d escdelay] [-h n] "
-		"[-s status-fifo] "
-		"[-c cmd-fifo] "
-		"[cmd...]\n");
+	eprint("usage: dvtm [-v] [-M] [-m mod] [-d delay] [-h lines] [-t title] "
+	       "[-s status-fifo] [-c cmd-fifo] [cmd...]\n");
 	exit(EXIT_FAILURE);
 }
 
@@ -1235,12 +1288,15 @@ parse_args(int argc, char *argv[]) {
 			create(args);
 			continue;
 		}
-		if (argv[arg][1] != 'v' && (arg + 1) >= argc)
+		if (argv[arg][1] != 'v' && argv[arg][1] != 'M' && (arg + 1) >= argc)
 			usage();
 		switch (argv[arg][1]) {
 			case 'v':
-				puts("dvtm-"VERSION" (c) 2007-2012 Marc Andre Tanner");
+				puts("dvtm-"VERSION" © 2007-2013 Marc André Tanner");
 				exit(EXIT_SUCCESS);
+			case 'M':
+				mouse_events_enabled = !mouse_events_enabled;
+				break;
 			case 'm': {
 				char *mod = argv[++arg];
 				if (mod[0] == '^' && mod[1])
@@ -1259,6 +1315,9 @@ parse_args(int argc, char *argv[]) {
 			case 'h':
 				screen.history = atoi(argv[++arg]);
 				break;
+			case 't':
+				title = argv[++arg];
+				break;
 			case 's':
 				bar.fd = open_or_create_fifo(argv[++arg], &bar.file);
 				updatebarpos();
@@ -1346,6 +1405,9 @@ main(int argc, char *argv[]) {
 					}
 				} else if ((key = keybinding(0, code))) {
 					key->action.cmd(key->action.args);
+				} else if (sel && vt_copymode(sel->term)) {
+					vt_copymode_keypress(sel->term, code);
+					draw(sel);
 				} else {
 					keypress(code);
 				}
@@ -1361,7 +1423,7 @@ main(int argc, char *argv[]) {
 			handle_statusbar();
 
 		for (c = clients; c; ) {
-			if (FD_ISSET(c->pty, &rd)) {
+			if (FD_ISSET(c->pty, &rd) && !vt_copymode(c->term)) {
 				if (vt_process(c->term) < 0 && errno == EIO) {
 					/* client probably terminated */
 					t = c->next;
